
#ifdef __APPLE__

/*
 * This is an ugly hack to circumvent a bug in Accelerate's cblas_sgemv.
 * Unfortunately I cannot know if we are using Accelerate Framework
 * so users of MKL, OpenBLAS and ATLAS will also get a performace
 * reduction from this patch. It should be removed as soon as Apple
 * fixes the bug in Accelerate.
 *
 * See:
 * https://github.com/numpy/numpy/issues/4007#issuecomment-45541678
 *
 * 
 */ 
 
#define NPY_NO_DEPRECATED_API NPY_API_VERSION
#include "Python.h"

#include "numpy/arrayobject.h" 
#include "npy_config.h"
#include "npy_pycompat.h"
#ifndef CBLAS_HEADER
#define CBLAS_HEADER "cblas.h"
#endif
#include CBLAS_HEADER

#include <string.h>



static void *aligned_malloc(npy_intp size, npy_intp align)
/* 
 * Aligned malloc on the Python heap
 * This function requires the GIL.
 *
 */
{
	void *p1, **p2;
	npy_intp offs = align - 1 + sizeof(void*);
	p1 = (void*)PyMem_Malloc(size + offs);
	if (p1 == NULL) return PyErr_NoMemory();
	p2 = (void**)(((npy_intp)(p1)+offs) & ~(align-1));
	*(p2-1) = p1;
	return (void*)p2;
}


static void aligned_free(void *p)
/* 
 * Free aligned buffer from the Python heap.
 * This function requires the GIL.
 *
 */
{
    
	PyMem_Free(*(((void**)p)-1));
} 
 
   
   
static float *aligned_matrix(const enum CBLAS_ORDER order, 
                    const npy_intp m, const npy_intp n, const float *A, 
                    const npy_intp lda)
/* 
 * Return an aligned copy of matrix A
 * if it is misaligned to 32 byte boundary. 
 * This function requires the GIL.
 *
 */
{
       
    float *alignedA;
    npy_intp r, c;
    if (((npy_intp)(void*)A)%32) {
        npy_intp sizeA = (order == CblasRowMajor ? m * lda : lda * n);          
        alignedA = (float*)aligned_malloc(m*n*sizeof(float),32);
        if(alignedA == NULL) return (float*)NULL;       
        if (order == CblasRowMajor) {
            for (r=0; r<m; r++) 
                memcpy((void*)(alignedA+r*n),(void*)(A+r*lda),n*sizeof(float));
        } else {
            for (c=0; c<n; r++) 
                memcpy((void*)(alignedA+c*m),(void*)(A+c*lda),m*sizeof(float));
        }        
    } else {
        alignedA = (float*)A;        
    }
    return alignedA;
}



static float *aligned_vector(const float *V, 
                             const npy_intp n, const npy_intp inc) 
/* 
 * Return an aligned copy of vector V
 * if it is misaligned to 32 byte boundary.
 * This function requires the GIL.
 *
 */ 
 {
    float *alignedV, *tmp;
    npy_intp i;
    if (((npy_intp)(void*)V)%32) {
        alignedV = (float*)aligned_malloc(n*sizeof(float),32);
        if(alignedV == NULL) return NULL;
        tmp = alignedV;
        for (i=0; i<n; i++) {
            *tmp++ = *V;
            V += inc;
        }
    } else {
        alignedV = (float*)V;
    }
    return alignedV;
}




/* ----------------------------------------------------------------- */

 
int patched_cblas_sgemv(const enum CBLAS_ORDER order, 
                        const enum CBLAS_TRANSPOSE trans,
                        const npy_intp m, npy_intp n, const float alpha,
                        const float *A, const npy_intp lda, 
                        const float *B, const npy_intp incB,
                        const float beta,
                        float *C, const npy_intp incC)
/*
 * Patch for the cblas_sgemv segfault in Accelerate: 
 * Aligns all input arrays on 32 byte boundaries.     
 *
 * This function can be called without owning the GIL.
 *
 *
 */
{
    float *alignedA = (float*)A, *alignedB = (float*)B, *alignedC = (float*)C;
    int freeA=0, freeB=0, freeC=0;
    npy_intp veclen = (trans == CblasTrans ? m : n);
    int _incB = incB, _incC = incC, _lda = lda;
        
    const int misaligned_arrays = (((npy_intp)(void*)A)%32) 
          || (((npy_intp)(void*)B)%32)
          || (((npy_intp)(void*)C)%32);
    
    /* 
     * Make temporary copies of misaligned arrays. 
     *
     */
    
    if (misaligned_arrays) {
    
        /* Ensure we have the GIL back for PyMem_* API */
        PyGILState_STATE state = PyGILState_Ensure();     
                
        /* aligned copy of A if needed */
        alignedA = aligned_matrix(order, m, n, A, lda);
        if (alignedA == NULL) {
            PyGILState_Release(state);
            return 1;
        }  
        if (alignedA != A) {
            _lda = (order == CblasRowMajor ? n : m);
            freeA = 1;
        }    
        
        /* aligned copy of B if needed */
        alignedB = aligned_vector(B, veclen, incB);
        if (alignedB == NULL) {
            aligned_free(alignedA);
            PyGILState_Release(state);
            return 1;
        }
        if (alignedB != B) {
            _incB = 1;
            freeB = 1;
        }       
        
        /* aligned copy of C if needed */
        alignedC = aligned_vector(C, veclen, incC);
        if (alignedC == NULL) {
            aligned_free(alignedB);
            aligned_free(alignedA);
            PyGILState_Release(state);
            return 1;
        }
        if (alignedC != C) {
            _incC = 1;
            freeC = 1;
        }  
        
        PyGILState_Release(state);
            
    }
        
    /* 
     * Now we can safely call cblas_sgemv. Arrays are aligned
     * to 32 byte boundaries and the GIL is in the same 
     * state as _dotblas left it.
     *
     */
    
    cblas_sgemv(order, trans, m, n, alpha, alignedA, _lda, 
                alignedB, _incB, beta, alignedC, _incC);
                        
                                
    /* 
     * Clean up temporary arrays if they were made.
     *
     */
    if (misaligned_arrays) {
          
        /* Ensure we have the GIL back for PyMem_* API */
        PyGILState_STATE state = PyGILState_Ensure();        
        if (freeA) aligned_free(alignedA);
        if (freeB) aligned_free(alignedB);
        if (freeC) {
            npy_intp i;
            float *tmp1 = C, *tmp2 = alignedC;
            for (i=0; i<veclen; i++) {
                *tmp1 = *tmp2;
                tmp1 += incC;
                tmp2 += _incC;
            }
            aligned_free(alignedC);
        }        
        PyGILState_Release(state);
    }
    
    return 0;

}

#endif

